from __future__ import annotations
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Callable, TypeVar, Any

registry: dict[str, FlexiDict] = {}

class FlexiDict(ABC):
    @abstractmethod
    def get_entry(self, n: int) -> str: ...

    @abstractmethod
    def preprocess(self, args: tuple[Any, ...], kwargs: dict[str, Any]) -> FlexiDict: ...

    @property
    @abstractmethod
    def size(self) -> int: ...

    @staticmethod
    def register(name: str, d: FlexiDict) -> None:
        registry[name] = d

    @staticmethod
    def get_registry() -> dict[str, FlexiDict]:
        return registry

T = TypeVar("T", bound=type[FlexiDict])

def register_dict(name: str) -> Callable[[T], T]:
    def decorator(cls: T) -> T:
        FlexiDict.register(name, cls())
        return cls

    return decorator

class FlexiTextDict(FlexiDict):
    def __init__(self, name: str, words: list[str], min: int = 0, max: int = 2048) -> None:
        if min != 0 or max != 2048:
            self.words = [w for w in words if len(w) >= min and len(w) <= max]
        else:
            self.words = words
        self.sz = len(self.words)
        self.name = name

    def get_entry(self, n: int) -> str:
        return self.words[n]

    def preprocess(self, args: tuple[Any, ...], kwargs: dict[str, Any]) -> FlexiTextDict:
        return FlexiTextDict(self.name, self.words, *args, **kwargs)

    @property
    def size(self) -> int:
        return self.sz

    @staticmethod
    def from_file(name: str, path: str) -> FlexiTextDict:
        fullpath = Path(path).resolve()
        with open(fullpath, "r") as file:
            lines = file.read().splitlines()
        d = FlexiTextDict(name, lines)
        FlexiDict.register(name, d)
        return d

@register_dict("hex")
class FlexiHexDict(FlexiDict):
    def __init__(self, size: int = 4) -> None:
        self.sz = size

    def get_entry(self, n: int) -> str:
        return f"{n:0{self.sz}x}"

    def preprocess(self, args: tuple[Any, ...], kwargs: dict[str, Any]) -> FlexiHexDict:
        return FlexiHexDict(*args, **kwargs)

    @property
    def size(self) -> int:
        num_bits = self.sz * 4
        ret: int = 2 ** num_bits
        return ret

@register_dict("decimal")
class FlexiDecimalDict(FlexiDict):
    def __init__(self, size: int = 4) -> None:
        self.sz = size

    def get_entry(self, n: int) -> str:
        return f"{n:0{self.sz}d}" 

    @property
    def size(self) -> int:
        ret: int = 10 ** self.sz
        return ret

    def preprocess(self, args: tuple[Any, ...], kwargs: dict[str, Any]) -> FlexiDecimalDict:
        return FlexiDecimalDict(*args, **kwargs)

FlexiTextDict.from_file("noun", "data/build/noun")
FlexiTextDict.from_file("adj", "data/build/adjective")
FlexiTextDict.from_file("verb", "data/build/verb")
FlexiTextDict.from_file("city", "data/build/city")
FlexiTextDict.from_file("firstname", "data/build/first-name")
FlexiTextDict.from_file("lastname", "data/build/last-name")
FlexiTextDict.from_file("femalename", "data/build/female-name")
FlexiTextDict.from_file("malename", "data/build/male-name")



from __future__ import annotations

import uuid

class RandomSource:
    def __init__(self, b: bytes) -> None:
        # print(f"creating random source with {len(b)} bytes")
        self.bytes = b
        self.big_num = int.from_bytes(self.bytes, byteorder="big")

    def get_max(self, max: int) -> int:
        # print(f"starting: {self.big_num}")
        self.big_num, res = divmod(self.big_num, max)
        # print(f"max: {max}, res: {res}, remaining: {self.big_num}")
        return res

    @staticmethod
    def from_uuid(u: uuid.UUID | str) -> RandomSource:
        if isinstance(u, str):
            u = uuid.UUID(u)
        return RandomSource(u.bytes)

    # from_hash(alg: str, alg_opts)
    # from_rand(nbytes: int = 16)



from __future__ import annotations

from jinja2 import Environment, BaseLoader
from typing import Any
import os
import hashlib
import math

from .dict import FlexiDict
from .rand import RandomSource


class FlexiHumanHash:
    def __init__(self, template: str) -> None:
        self.template_str = template
        self.jinja_env: Environment = Environment(loader=BaseLoader)  # type: ignore
        dicts = FlexiDict.get_registry()
        for dict_name in dicts.keys():
            self.jinja_env.globals[dict_name] = JinjaExpr(self, dict_name, dicts[dict_name])
        self.jinja_template = self.jinja_env.from_string(self.template_str)
        self.call_records: list[JinjaExpr] = []
        self.rndctx: RandomSource | None = None
        # XXX: when rndctx is None rendering the template has side-effects recording calls to calculate entropy
        self.jinja_template.render()
        entropy = 1
        num_bits = 0
        for r in self.call_records:
            entropy *= r.dict.size
            num_bits += required_bits(r.dict.size)
        self.entropy = entropy
        self.entropy_bits = num_bits
        self.required_bytes = required_bytes(self.entropy_bits)

    def hash(self, data: bytes | int | str, alg: str = "blake2b") -> HashResult:
        input = data
        if isinstance(data, int):
            sz = required_bytes(required_bits(data))
            sz = sz if sz > 4 else 4
            data = data.to_bytes(sz)

        if isinstance(data, str):
            data = data.encode()

        match alg:
            case "blake2b":
                data = hashlib.blake2b(data, digest_size=self.required_bytes).digest()
                source = f"hash:{alg}:{self.required_bytes}"
            case "blake2s":
                data = hashlib.blake2s(data, digest_size=self.required_bytes).digest()
                source = f"hash:{alg}:{self.required_bytes}"
            case "shake128":
                data = hashlib.shake_128(data).digest(self.required_bytes)
                source = f"hash:{alg}:{self.required_bytes}"
            case "shake256":
                data = hashlib.shake_256(data).digest(self.required_bytes)
                source = f"hash:{alg}:{self.required_bytes}"
            case "md5":
                data = hashlib.md5(data).digest()
                source = f"hash:{alg}"
            case "sha1":
                data = hashlib.sha1(data).digest()
                source = f"hash:{alg}"
            case "sha224":
                data = hashlib.sha224(data).digest()
                source = f"hash:{alg}"
            case "sha256":
                data = hashlib.sha256(data).digest()
                source = f"hash:{alg}"
            case "sha384":
                data = hashlib.sha384(data).digest()
                source = f"hash:{alg}"
            case "sha512":
                data = hashlib.sha512(data).digest()
                source = f"hash:{alg}"
            case "sha3-224":
                data = hashlib.sha3_224(data).digest()
                source = f"hash:{alg}"
            case "sha3-256":
                data = hashlib.sha3_256(data).digest()
                source = f"hash:{alg}"
            case "sha3-384":
                data = hashlib.sha3_384(data).digest()
                source = f"hash:{alg}"
            case "sha3-512":
                data = hashlib.sha3_512(data).digest()
                source = f"hash:{alg}"
            case _:
                raise TypeError(f"unknown hash algorithm: {alg}")

        res = self.from_bytes(data)
        res.input_source = source
        res.input = input
        return res

    # def hash_str(self, data: str) -> HashResult:
    #     return HashResult("")

    # from_hexstr

    # def hash_int(self, data: int) -> HashResult:
    #     return HashResult("")

    # def from_uuid

    def rand(self) -> HashResult:
        data = os.urandom(self.required_bytes)

        ret = self.from_bytes(data)
        ret.input_source = "rand"
        return ret

    def from_bytes(self, data: bytes) -> HashResult:
        input_data = data
        if len(data) < self.required_bytes:
            need_bytes = self.required_bytes - len(data)
            data = data + os.urandom(need_bytes)

        self.rndctx = RandomSource(data)
        ret = self.jinja_template.render()
        self.rndctx = None

        return HashResult(
            hasher=self,
            result=ret,
            input_source="bytes",
            input=input_data,
            input_data=input_data,
            data=data,
            bits_provided=len(input_data) * 8,
            bits_used=self.entropy_bits,
            entropy=self.entropy,
        )


class HashResult:
    def __init__(
        self,
        *,
        hasher: FlexiHumanHash,
        result: str,
        input: Any,
        input_data: bytes,
        data: bytes,
        bits_provided: int,
        input_source: str,
        bits_used: int,
        entropy: int,
    ) -> None:
        self.result = result
        self.input = input
        self.input_data = input_data
        self.data = data
        self.input_source = input_source
        self.bits_provided = bits_provided
        self.bits_used = bits_used
        self.entropy = entropy
        self.hasher = hasher

    def __str__(self) -> str:
        return self.result


class JinjaExpr:
    def __init__(self, hasher: FlexiHumanHash, name: str, flexi_dict: FlexiDict) -> None:
        self.hasher = hasher
        self.name = name
        self.dict = flexi_dict
        self.args: tuple[Any, ...] | None = None
        self.kwargs: dict[str, Any] | None = None

    def __call__(self, *args: Any, **kwargs: dict[str, Any]) -> str:
        if self.hasher.rndctx is None:
            self.preprocess(args, kwargs)
            return ""

        return self.get_word()

    def __str__(self) -> str:
        if self.hasher.rndctx is None:
            self.preprocess()
            return ""

        return self.get_word()

    def get_word(self) -> str:
        r = self.hasher.rndctx
        d = self.dict
        assert r is not None
        idx = r.get_max(d.size)
        return d.get_entry(idx)

    def preprocess(
        self,
        args: tuple[Any, ...] = tuple(),
        kwargs: dict[str, Any] = dict(),
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.hasher.call_records.append(self)
        self.dict = self.dict.preprocess(args, kwargs)


def required_bits(n: int) -> int:
    return math.ceil(math.log2(n))


def required_bytes(bits: int) -> int:
    return math.ceil(bits / 8)
